from os import getenv
from os.path import dirname, abspath, isabs, join
import yaml
from fabric.api import env

def mkabs(home, path):
    return path if isabs(path) else join(home, path)

def fabex_config(config=None, config_file=None):
    """Use config or fabex config file, hand it off env.fabex_config"""
    if config:
        env.fabex_config = config
    else:
        config_file = config_file or getenv('FABEX_CONFIG', 'fabex_config.yaml')
        with open(config_file) as f:
            env.fabex_config = yaml.load(f)

    home = env.fabex_config['fabfile_home'] = dirname(env.real_fabfile)
    env.fabex_config['target_dir'] = env.fabex_config.get('target_dir', 'targets')
    env.fabex_config['template_dir'] = env.fabex_config.get('template_dir', 'templates')

    # if env.roles set from cl or top of fabfile we populate env.roledefs, so
    # role based task selection
    if env.roles:
        env.roledefs = { role: list() for role in env.roles }
    if 'template_config' in env.fabex_config:
        with open(env.fabex_config['template_config']) as f:
            env.fabex_templates = yaml.load(f)
    else:
        env.fabex_templates = {}



from __future__ import unicode_literals

from os.path import isabs, join
from fabric.api import env
from fabric.contrib.files import *
from fabric.contrib.files import (exists as fab_exists,
                                  upload_template as fab_upload_template,
                                  sed as fab_sed,
                                  append as fab_append,
                                  )

from fabric.operations import _AttributeList # to mimic in dryruns
from jinja2 import Environment
from ..fab_api import print_command, sudo, run
from ..exceptions import FabexConfigError, FabexRunError

__all__ = ['exists', 'upload_template', 'sed', 'append', 'upload_project_template',
           'rm_project_template']

def exists(path, **kwargs):
    __doc__ = fab_exists.__doc__

    if env.dryrun:
        return False
    return fab_exists(path, **kwargs)


def upload_template(filename, destination, **kwargs):
    __doc__ = fab_exists.__doc__

    if env.dryrun:
        print_command("upload_template {} -> {}".format(filename, destination))

        return _AttributeList(destination)
    return fab_upload_template(filename, destination, **kwargs)


def sed(filename, before, after, **kwargs):
    __doc__ = fab_sed.__doc__
    if env.dryrun:
        args = {'backup': kwargs.get('backup', '.bak'),
                'limit': kwargs.get('limit', ''),
                'flags': kwargs.get('flags', ''),
                'before': before,
                'after': after,
                'filename': filename,
               }
        print_command('sed -i{backup} -r -e "/{limit}/ '
                      's/{before}/{after}/{flags}g {filename}"'
                      .format(**args))
        return "[sed output]"
    return fab_sed(filename, before, after, **kwargs)


def append(filename, text, **kwargs):
    __doc__ = fab_append.__doc__
    if env.dryrun:
        print_command("append {}->{}".format(filename, text))
        return
    return fab_append(filename, text, **kwargs)



def jinja2_string(template, context=env):
    return Environment().from_string(template).render(**context)


def get_templates(templates=[]):
    """Returns each of the templates with env vars injected."""
    injected = {}
    for name, data in (t for t in env.fabex_templates.iteritems()
                       if not templates or t[0] in templates):
        injected[name] = {k: jinja2_string(v) for k, v in data.items()}
    return injected


def upload_project_template(name, force=False, reload=True):
    """Uploads a template if changed. Execute reload_command if changed and reload. """
    try:
        template = get_templates([name])[name]
    except KeyError:
        raise FabexConfigError("missing template: {}".format(name))
    local_path = template["local_path"]
    if not isabs(local_path):
        local_path = join(env.fabex_config['template_dir'], local_path)
    if not os.path.exists(local_path):
        raise FabexRunError("template {} not found at {}".format(name, local_path))
    remote_path = template["remote_path"]
    reload_command = template.get("reload_command")
    owner = template.get("owner")
    mode = template.get("mode")
    remote_data = ""
    if exists(remote_path):
        with hide("stdout"):
            remote_data = sudo("cat %s" % remote_path, quiet=True, shell=False)
    with open(local_path, "r") as f:
        local_data = jinja2_string(f.read())
    whitewash = lambda s: re.sub(r"(?m)\s+", "", s)
    if whitewash(remote_data) == whitewash(local_data):
        if force:
            print_command("Template {} unchanged, but force=True".format(name), arrow_out=False)
        else:
            print_command("Template {} unchanged, not uploading".format(name), arrow_out=False)
            return
    upload_template(local_path, remote_path, context=env,
                    use_sudo=True, backup=False, use_jinja=True)
    if owner:
        sudo("chown %s %s" % (owner, remote_path))
    if mode:
        sudo("chmod %s %s" % (mode, remote_path))
    if reload and reload_command:
        return sudo(reload_command)
    return ""


def rm_project_template(name, reload=True, force=False):
    """Removes a template on the remote host"""
    try:
        template = get_templates([name])[name]
    except KeyError:
        raise FabexConfigError("missing template: {}".format(name))
    remote_path = template["remote_path"]
    reload_command = template.get("reload_command")
    if not exists(remote_path):
        if force:
            return ""
        raise FabexRunError("template {} does not exist at remote path {}"
                            .format(name, remote_path))
    sudo("rm -v -f " + remote_path)
    if reload and reload_command:
        return sudo(reload_command)
    return ""





from __future__ import absolute_import

from fabric.api import *
from fabex.config import fabex_config
from fabex.fab_api import run, sudo, cd, prefix, print_command
from fabex.tasks import dryrun, verbose, target, password, encrypt
from fabex.decorators import task_roles, runs_once_per_host
from fabex.context_managers import role_settings, host_settings



import yaml
from os.path import isabs, join
from collections import defaultdict
from fabric.api import env, task, runs_once, require, execute
from fabric.tasks import WrappedCallableTask, Task
from fabric import state

from fabex.exceptions import FabexConfigError
from fabex.context_managers import role_settings, host_settings
from .crypt import AESCipher


__all__ = ['dryrun', 'verbose', 'target', 'roles', 'password', 'encrypt']

env.dryrun = False
env.verbose = True

__cipher = None

@task
@runs_once
def dryrun(dryrun="yes"):
    """Fabex: Set dryrun for no execution of commands"""
    env.dryrun = dryrun.lower() in ('y', 'yes', 't', 'true')


@task
@runs_once
def verbose(verbose="yes"):
    """Fabex: Shows running and output of tasks if =yes"""
    env.verbose = verbose.lower() in ('y', 'yes', 't', 'true')


@task
@runs_once
def quiet(quiet="yes"):
    """Fabex: Hides running and output of tasks if =yes"""
    env.verbose = not quiet.lower() in ('y', 'yes', 't', 'true')

@task
@runs_once
def password(pw):
    """Fabex: password:pw, set the encryption password for encrypt and yaml !decrypt"""
    global __cipher
    __cipher = AESCipher(pw)

@task
@runs_once
def encrypt(text):
    """Fabex: encrypt:text, must call password task first"""

    if not __cipher:
        raise FabexConfigError("password:pw required for encrypt task")
    print(__cipher.encrypt(text))

def decrypt_constructor(loader, node):
    """for !decrypt('encrypted-string') in yaml"""
    if not __cipher:
        raise yaml.YAMLError("target uses !decrypt but password:pw not called")
    etext = loader.construct_scalar(node)
    text = __cipher.decrypt(etext)
    if not text:
        raise yaml.YAMLError("{} could not be decrypted".format(etext))
    return text

yaml.add_constructor(u'!decrypt', decrypt_constructor)

@task
@runs_once
def target(name):
    """Fabex: target:name, read the name.yaml configuration"""

    require('fabex_config')

    target_path = name
    if not isabs(target_path) and 'target_dir' in env.fabex_config:
        target_path = join(env.fabex_config['target_dir'], target_path)
    if not target_path.endswith('.yaml'):
        target_path += '.yaml'
    with open(target_path) as f:
        env.update(yaml.load(f))
    env.target = name

    hostroles = defaultdict(set)
    for role, hosts in env.roledefs.items():
        for host in hosts:
            hostroles[host].add(role)
    env.hostroles = dict(hostroles)

    # Alongside env.hosts is env.roles (not to be confused with env.roledefs!)
    # which, if given, will be taken as a list of role names to look up in env.roledefs.
    #
    # We accept a preset env.roles list, from command line or otherwise before this task
    # executes. However, an empty roles list means all roles from target role_settings.
    # Also, assert that all roles in use have an entry in env.roledefs.
    if not set(env.roles).issubset(env.roledefs.keys()):
        raise FabexConfigError("roles in {} not found in roledefs: {}"
                               .format(','.join(env.roles), ','.join(env.roledefs.keys())))

    # Handle 'extends' in role_settings
    for role, settings in env.roleenvs.items():
        extends = settings.get('extends')
        if not extends:
            continue
        extends = extends if isinstance(extends, list) else [extends]
        settings_backup = settings.copy()
        for parent in extends:
            if parent not in env.roledefs:
                raise FabexConfigError("extends refers to nonexistant role: "
                                       "{} in role_settings for {}".format(parent, role))
            settings.update(env.roleenvs.get(parent, {}))
        # child settings always override parent
        settings.update(settings_backup)

    env.verbose = env.get('verbose', True)


class TaskGroup(Task):
    def __init__(self, name, *args, **kwargs):
        self.tasks = []
        self.roles = []
        self.__doc__ = "Fabex group: ".format(name)
        super(TaskGroup, self).__init__(name=name, *args, **kwargs)

    def append(self, task):
        if not env.roles or set(task.wrapped.roles).intersection(env.roles):
            # don't add to group if --roles= and task does not have requested role
            self.tasks.append(task)
            self.__doc__ += ("" if self.__doc__.endswith(" ") else ", ") + task.name
            self.roles.extend([r for r in task.wrapped.roles if r not in self.roles])

    def run(self, *args, **kwargs):
        task_role_filter = lambda t: env.hostroles[env.host].intersection(t.wrapped.roles)
        task_filter = ((lambda t: task_role_filter(t).intersection(env.roles)) if env.roles
                       else task_role_filter)
        tasks = filter(task_filter, self.tasks)
        return {task.name: execute(task, *args, host=env.host, **kwargs) for task in tasks}

    def __repr__(self):
        return "TaskGroup({},[{}])".format(self.name, ','.join(map(str, self.tasks)))


class RoleTask(WrappedCallableTask):
    def run(self, *args, **kwargs):
        ret = {}
        roles_requested = env.roles or env.hostroles[env.host]
        roles = filter(lambda r: r in roles_requested and env.host in env.roledefs[r], self.roles)
        for role in roles:
            with role_settings(role), host_settings(env.host):
                if state.output.running and not hasattr(self, 'return_value'):
                    print("[{host}:{role}] Executing role task '{command}'".format(**env))
                ret[role] = super(RoleTask, self).run(*args, **kwargs)
        return ret

    def __repr__(self):
        return "RoleTask({},[{}])".format(self.name, ','.join(self.roles))



from __future__ import absolute_import

from contextlib import contextmanager
from fabric.api import env, settings
from .exceptions import FabexConfigError

@contextmanager
def role_settings(role, **kwargs):
    """Inject env.roleenvs[role], plus settings from kwargs"""

    new_settings = kwargs.copy()
    roledict = env.roleenvs.get(role)
    if not roledict:
        raise FabexConfigError("role {} not found in target roleenvs".format(role))
    new_settings.update(roledict)
    new_settings['role'] = role
    with settings(**new_settings):
        yield

@contextmanager
def host_settings(host, **kwargs):
    """Inject env.hostenvs[role], plus settings from kwargs"""

    new_settings = kwargs.copy()
    hostdict = env.hostenvs.get(host)
    if not hostdict:
        raise FabexConfigError("host {} not found in target hostenvs".format(host))
    new_settings.update(hostdict)
    with settings(**new_settings):
        env.host_string = env.get('ssh_host', env.host)
        if 'ssh_user' in env:
            env.host_string = env['ssh_user'] + '@' + env.host_string
        yield



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='Fabex',
    version='1.1',
    description='Python Fabric Extended: Making roles more good',
    long_description=README, 
    author='Rod Morison',
    author_email='rod@rodmtech.net',
    url='http://fabex.readthedocs.org/en/latest/',
    packages=['fabex', 'fabex.contrib'],
    scripts=['scripts/fxcrypt'],
    install_requires=['fabric', 'jinja2', 'PyYAML', 'pycrypto'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
    ]
)



# -*- coding: utf-8 -*-
"""Facade over fabric.api, for env.dryrun support, and printing run/sudo commands"""
from __future__ import unicode_literals

import os
import sys
import re
import yaml
from contextlib import contextmanager
from fabric import colors
from fabric.api import *
# fabex overrides
from fabric.api import (run as fab_run, sudo as fab_sudo, cd as fab_cd, prefix as fab_prefix)

__all__ = ['print_command', 'run', 'sudo', 'cd', 'prefix']

_prefix_stack = []
_cd_stack = []

# if tty, use colors
if sys.stdout.isatty():
    red, green, blue, magenta = colors.red, colors.green, colors.blue, colors.magenta
else:
    red = green = blue = magenta = lambda x: x

def print_command(command, arrow_out=True):
    """Loudly display command with colors"""

    arrow = " ->" if arrow_out else " <-"
    pre = []
    if _cd_stack:
        pre.append(_cd_stack[-1])
    if _prefix_stack:
        pre.extend(_prefix_stack)
    if pre:
        pre.append('')
    pres = ' && '.join(pre)
    if 'role' in env:
        print("[{}:{}] $ {}{}{}".format(blue(env.host), blue(env.role), green(pres),
                                        green(command), arrow))
    else:
        print("[{}] $ {}{}{}".format(blue(env.host), green(pres), green(command), arrow))

def run(command, **kwargs):
    __doc__ = fab_run.__doc__

    if env.dryrun:
        print_command(command)
        return "[{} output]".format(command)
    to_hide = [] if env.verbose else ['running', 'output']
    with hide(*to_hide):
        return fab_run(command, **kwargs)

def sudo(command, **kwargs):
    __doc__ = fab_sudo.__doc__

    if env.dryrun:
        print_command(command)
        return "[{} output]".format(command)
    to_hide = [] if env.verbose else ['running', 'output']
    with hide(*to_hide):
        return fab_sudo(command, **kwargs)

@contextmanager
def prefix(command):
    __doc__ = fab_prefix.__doc__
    if env.dryrun:
        _prefix_stack.append(command)
        yield
        _prefix_stack.pop()
    else:
        with fab_prefix(command):
            yield

@contextmanager
def cd(path):
    __doc__ = fab_cd.__doc__
    if env.dryrun:
        _cd_stack.append(path)
        yield
        _cd_stack.pop()
    else:
        with fab_cd(path):
            yield




from __future__ import absolute_import

from sys import modules
from os.path import basename, splitext
from functools import wraps
from collections import defaultdict
from fabric.api import task, roles, env
from fabric.decorators import serial, _wrap_as_new

from fabex.exceptions import FabexRunError
from fabex.tasks import RoleTask, TaskGroup


def _get_fabmodule():
    try:
        return modules[splitext(basename(env.fabfile))[0]]
    except Exception as e:
        raise FabexRunError("error locating fabfile {}: {}".format(env.fabfile, e))


def task_roles(*args, **kwargs):
    invoked = bool(kwargs)
    roles_attach = roles(*args)
    group = kwargs.pop('group', None)
    def attach(func):
        task_wrapper = task(task_class=RoleTask, **kwargs)
        task_ = task_wrapper(roles_attach(func))
        if group:
            fabmodule = _get_fabmodule()
            if not hasattr(fabmodule, group):
                setattr(fabmodule, group, TaskGroup(group))
            getattr(fabmodule, group).append(task_)
        return task_
    return attach

_host_func_return_value = {}

def runs_once_per_host(func):
    @wraps(func)
    def decorated(*args, **kwargs):
        if (env.host, func) in _host_func_return_value:
            role = ":{role}".format(**env) if env.role else ""
            print("[{host}{}] Already executed '{command}' on this host".format(role, **env))
        else:
            _host_func_return_value[(env.host, func)] = func(*args, **kwargs)
        return _host_func_return_value[(env.host, func)]
    decorated = _wrap_as_new(func, decorated)
    # Mark as serial (disables parallelism) and return
    return serial(decorated)
